#!/usr/bin/perl
# -*- project-name: VASM: mode: cperl; mode: outline-minor -*-

### * Preamble
use strict;
use warnings;
use constant +{ 
  FALSE => 0, TRUE => 1,
  LABEL => 0, 
  LABEL_REAL => 0, LABEL_SHOWN => 1,
  PAD => 2 
};
use VASM::Tree;
use VASM::Resource;
use VASM::Resource::Catalog::Message;
use VASM::Resource::Catalog::Action;
use VASM::Utility::Gtk2::MessageDialog;
use Gtk2 '-init';
use File::Spec::Functions qw/curdir rel2abs splitpath/;
use File::MimeInfo::Magic;
use Archive::Tar;

# Message files
my $UICommonMsg = findMessageCatalog(qw/Action UI Common Message/);
my $UIMsg = findMessageCatalog(qw/Action UI Gtk2 Message/);
my $ErrorMsg =findErrorCatalog(qw/Action UI Error Message/);

### * Action Tree
### ** Internal
### *** makeActionTree
sub makeActionTree {
  # Get directories for file and URI schemata
  my $fileDirectory = mediaDirectory('File');
  my $uriDirectory = mediaDirectory('URI');

  # Convert the file directory to data more readily usable by the function
  my $fileTypeTree = VASM::Tree->new;
  for my $mime ($fileDirectory->List) {
    $fileTypeTree->Store((split m!/!, $mime), TRUE);
  }

  # Now, set up a tree store corresponding to them and add the 'Template'
  # top-level choice as well. The first Glib::String is the item as it appears
  # on the screen, according to translation and additional notes. The second
  # is the 'real name', usable to the program.
  my $actionTree = Gtk2::TreeStore->new('Glib::String', 'Glib::String');
  my @iters;

  # File
  push @iters, $actionTree->append(undef);
  $actionTree->set(
    $iters[-1],
    LABEL_REAL, 'File',
    LABEL_SHOWN, $UICommonMsg->Render('File')
  );
  for my $class (sort $fileTypeTree->Children) {
    # Add the class under 'File'
    push @iters, $actionTree->append($iters[-1]);
    $actionTree->set(
      $iters[-1],
      LABEL_REAL, $class,
      LABEL_SHOWN, $class
    );
    # Add each member of the class (e.g., File -> image -> jpeg)
    for my $member (sort $fileTypeTree->Children($class)) {
      push @iters, $actionTree->append($iters[-1]);
      my $mime = join('/', $class, $member);
      # The type will appear with a parenthetical note describing it in
      # further detail
      $actionTree->set(
        $iters[-1],
        LABEL_REAL, $member,
        LABEL_SHOWN, $member . ' (' . $fileDirectory->Render($mime) . ')'
      );
    } continue { pop @iters }
  } continue { pop @iters }
  pop @iters;

  # URI
  push @iters, $actionTree->append(undef);
  $actionTree->set(
    $iters[-1],
    LABEL_REAL, 'URI',
    LABEL_SHOWN, $UICommonMsg->Render('URI')
  );

  for my $scheme (sort $uriDirectory->List) {
    push @iters, $actionTree->append($iters[-1]);
    $actionTree->set(
      $iters[-1],
      LABEL_REAL, $scheme,
      LABEL_SHOWN, $scheme . ' (' . $uriDirectory->Render($scheme) . ')'
    );
  } continue { pop @iters }
  pop @iters;

  # Template
  push @iters, $actionTree->append(undef);
  $actionTree->set(
    $iters[-1],
    LABEL_REAL, 'Template',
    LABEL_SHOWN, $UICommonMsg->Render('Template')
  );
  pop @iters; # Neat freak

  return $actionTree;
}

### *** makeActionTreeView
sub makeActionTreeView {
  my $actionTree = makeActionTree(); # The actual data
  # The rest of this is a 'cookbook' recipe to make a tree view correspond
  # one-for-one with the data in the tree store.
  my $actionTreeView = Gtk2::TreeView->new_with_model($actionTree);
  $actionTreeView->set_headers_visible(FALSE);
  my $col = Gtk2::TreeViewColumn->new;
  my $renderer = Gtk2::CellRendererText->new;
  $col->pack_start($renderer, FALSE);
  $col->add_attribute($renderer, text => LABEL_SHOWN);
  $actionTreeView->append_column($col);
  
  # Attach a signal handler for double clicks on rows
  $actionTreeView->signal_connect(
    'row-activated' => sub { 
      dispatchMediaType(identifyActionTreeRow($_[0]))
    });

  return $actionTreeView;
}

### *** identifyActionTreeRow
sub identifyActionTreeRow {
  my ($actionTree, $iter) = $_[0]->get_selection->get_selected; my @path;

  # Walk from branch to trunk, unshifting the 'real' column at each point. The
  # result will define the path in the tree store.
  do {
    unshift @path, $actionTree->get($iter, LABEL_REAL);
  } while ($iter = $actionTree->iter_parent($iter));

  return @path;
}

### ** External
### *** dispatchMediaType
sub dispatchMediaType {
  my @path = @_;

  # Fully qualified path was given
  if (($path[0] eq 'File' and @path == 3) or
      ($path[0] eq 'URI' and @path == 2)) {
    configureMedium(@path);
  } elsif ($path[0] eq 'Template') { # Select a template
    selectTemplate();
  }
  
  return;
}

### * Action List
### ** Internal
### *** makeActionList
sub makeActionList {
  my ($catalog) = @_; my $iter;
  my $actionList = Gtk2::ListStore->new('Glib::String');
  my @programs = $catalog->Dump;

  if (@programs) { # Add each program in the catalog to the list store
    for my $program (@programs) {
      $iter = $actionList->append;
      $actionList->set($iter, LABEL, $program);
    }
  } else { # Empty or non-existent catalog
    $iter = $actionList->append;
    $actionList->set($iter, LABEL, ''); # Add a new element
  }
  
  return $actionList;
}

### *** makeActionListView
sub makeActionListView {
  my ($catalog) = @_;

  # Cookbook recipe that instantiates a list view store from an actions
  # catalog and imposes an editable view component on it
  my $actionList = makeActionList($catalog);
  my $actionListView = Gtk2::TreeView->new_with_model($actionList);
  $actionListView->set_headers_visible(FALSE);
  # $view->set_reorderable(TRUE); # Allow drag and drop :D
  $actionListView->grab_focus; # Using keyboard shortcuts becomes easier
  my $col = Gtk2::TreeViewColumn->new;
  my $renderer = Gtk2::CellRendererText->new;
  $renderer->set(editable => TRUE);
  $renderer->signal_connect(edited => \&actionListCellEdited, $actionList);
  $col->pack_start($renderer, FALSE);
  $col->add_attribute($renderer, text => LABEL);
  $actionListView->append_column($col);

  return $actionListView;
}

### *** actionListFull
sub actionListFull {
  my ($actionList) = @_; my $iter = $actionList->get_iter_first;
  
  # Return FALSE unless all elements of the list store are not blank
  while (defined $iter) {
    return FALSE unless length $actionList->get($iter, LABEL);
    $iter = $actionList->iter_next($iter);
  }

  return TRUE; # Fall through
}

### ** External
### *** configureMedium
sub configureMedium {
  my @path = @_;
  my $mediaType = (@path == 3) ? join '/', @path[1..2] : $path[1];

  # First, load the action catalog, if possible
  my $catalog = findActionCatalog(@path);
  $catalog = VASM::Catalog::Action->new unless defined $catalog;

  # Dialog
  my $dialog = Gtk2::Dialog->new_with_buttons(
    $mediaType,
    undef, [],
    'gtk-ok' => 'ok',
    'gtk-cancel' => 'cancel');
  $dialog->set_position('center');

  # Media list
  my $actionListView = makeActionListView($catalog);
  $dialog->vbox->pack_start($actionListView, TRUE, TRUE, PAD);

  # Operations
  my $operationHButtonBox = Gtk2::HButtonBox->new;
  $operationHButtonBox->set_layout('spread');
  my @operations = ( [ 'gtk-add' => \&actionListAdd ],
                     [ 'gtk-remove' => \&actionListRemove ],
                     [ 'gtk-go-up' => \&actionListUp ],
                     [ 'gtk-go-down' => \&actionListDown ] );
  # Add them, one by one
  for my $operation (@operations) {
    my $button = Gtk2::Button->new_from_stock($operation->[0]);
    $button->signal_connect(clicked => $operation->[1], $actionListView);
    $operationHButtonBox->pack_start($button, FALSE, FALSE, PAD);
  }
  $dialog->vbox->pack_start($operationHButtonBox, FALSE, FALSE, PAD);

  # Description
  my $shortDescription = Gtk2::Expander->new(
    $UIMsg->Render(id => 'Medium Configurator Short Description',
                   arguments => [ $mediaType ]));
  my $longDescription = Gtk2::Label->new(
    $UIMsg->Render('Medium Configurator Long Description'));
  $longDescription->set_line_wrap(TRUE);
  $shortDescription->add($longDescription);
  $dialog->vbox->pack_start($shortDescription, FALSE, FALSE, PAD);

  $dialog->show_all;

  DIALOG: {
    my $response = $dialog->run;
    # Write out the catalog if so desired
    if ($response eq 'ok') {
      # Are all fields full?
      unless (actionListFull($actionListView->get_model)) {
        # Report error
        errorMessage($ErrorMsg->Render('Missing Fields'));
        redo DIALOG; # Do over
      }
      $catalog = generateCatalog($actionListView->get_model);
      writeActionCatalog($catalog, @path);
    }
  }
  $dialog->destroy;

  return;
}

### *** actionListCellEdited
sub actionListCellEdited {
  my ($pathString, $text, $actionList) = @_[1..3];

  # Just pass the text through
  $actionList->set($actionList->get_iter_from_string($pathString), 
                   LABEL, $text);
  
  return;
}

### *** actionListAdd
sub actionListAdd {
  my $actionListView = $_[1];
  my $actionList = $actionListView->get_model;
  my $iter = $actionList->insert_after(
    scalar $actionListView->get_selection->get_selected
  );
  $actionList->set($iter, LABEL, ''); # Add a new element

  return;
}

### *** actionListRemove
sub actionListRemove {
  my $actionListView = $_[1];
  my ($actionList, $iter) = $actionListView->get_selection->get_selected;
  $actionList->remove($iter) if $iter;

  return;
}

### *** actionListUp
sub actionListUp {
  my $actionListView = $_[1];
  my ($actionList, $iter) = $actionListView->get_selection->get_selected;
  my $prev = eval {
    my $path = $actionList->get_path($iter);
    return $actionList->get_iter($path) if $path->prev;
  };
  $actionList->swap($prev, $iter) if $prev;
  
  return;
}

### *** actionListDown
sub actionListDown {
  my $actionListView = $_[1];
  my ($actionList, $iter) = $actionListView->get_selection->get_selected;
  my $next = eval {
    my $path = $actionList->get_path($iter); $path->next;
    return $actionList->get_iter($path);
  };
  $actionList->swap($iter, $next) if $next;
  
  return;
}

### * Template
### ** External
### *** installTemplate
sub installTemplate {
  my ($templateFile) = @_;

  # Confirm the installation 
  my $response = confirmMessage(
    $UIMsg->Render('Confirm Installation')
  );
  return unless $response eq 'yes';

  # Check the MIME type of the file
  my $mimetype = magic($templateFile);

  # Raise an error if needed
  unless ($mimetype =~ m!^application/x-(?:compressed-)?tar$!) {
    errorMessage($ErrorMsg->Carp('Invalid Archive'));
    return;
  }

  # Find the resource path and create it if necessary
  my $root = makeConfigResourcePath('Action');

  # Set up the umask to produce -rw------- files and save the old one
  my $old_umask = umask; umask 0077;
  # Change directory and extract the template archive
  chdir $root; Archive::Tar->extract_archive($templateFile);
  # Revert to the old umask
  umask $old_umask;

  # Confirm the installation (to the user)
  infoMessage($UIMsg->Render('Installation Complete'));

  return;
}

### *** selectTemplate
BEGIN {
  my $wd = rel2abs(curdir);
  
  sub selectTemplate {
    my $fileChooser = Gtk2::FileChooserDialog->new(
      $UIMsg->Render('Select Template'), undef, 
      'open', 
      'gtk-ok' => 'ok',
      'gtk-cancel' => 'cancel');
    $fileChooser->set_current_folder($wd);
    $fileChooser->set_position('center');
    my $filter = Gtk2::FileFilter->new;
    # MIME-type selection seems unfortunately unreliable
    for my $ext (qw/tgz tar tar.gz/) {
       $filter->add_pattern("*.$ext");
    }
    $fileChooser->set_filter($filter);

    $fileChooser->show_all;
    my $response = $fileChooser->run;
    # If a given file is accepted...
    if ($response eq 'ok') {
      my $templateFile = $fileChooser->get_filename;
      # Confirm installation of the template and install if confirmed
      installTemplate($templateFile);
      # Store new current directory
      $wd = (splitpath($templateFile))[1];
    }
    $fileChooser->destroy;

    return;
  }
}

### * Main
### ** External
### *** mainMenu
sub mainMenu {
  # Dialog
  my $dialog = Gtk2::Dialog->new_with_buttons(
    $UIMsg->Render('Title'),
    undef, [],
    'gtk-close' => 'close');
  $dialog->set_position('center');
  $dialog->set_default_size(400, 300);
    
  ## Action area
  # Tree with media
  my $actionTreeView = makeActionTreeView();
  # Pack the tree view into a scrolled window. The addition of scrollbars
  # will be done automatically in either direction
  my $scrolledTree = Gtk2::ScrolledWindow->new;
  $scrolledTree->set_policy(qw/automatic/ x 2);
  $scrolledTree->add($actionTreeView);
  $dialog->vbox->pack_start($scrolledTree, TRUE, TRUE, PAD);
  # Expander with detailed description
  my $shortDescription = Gtk2::Expander->new(
    $UIMsg->Render('Main Menu Short Description')
  );
  my $longDescription = Gtk2::Label->new(
    $UIMsg->Render('Main Menu Long Description')
  );
  $longDescription->set_line_wrap(TRUE);
  $shortDescription->add($longDescription);
  $dialog->vbox->pack_start($shortDescription, FALSE, FALSE, PAD);

  # Do it Jah!
  $dialog->show_all;
  $dialog->run;
  $dialog->destroy;
  
  exit 0;
}

### *** generateCatalog
sub generateCatalog {
  my ($actionList) = @_; my $iter;
  my $catalog = VASM::Catalog::Action->new;

  # Add an element in the catalog for each member in the list store
  $actionList->foreach(
    sub { $catalog->Push($_[0]->get($_[2], LABEL)); return FALSE }
  );
  
  return $catalog;
}

### * Invocation
mainMenu();
